import sys
import getopt
import collections

from .exception import SftpException, SftpConfigException, SftpMountException
from .model import EnvironmentModel, SystemModel, SystemControllerModel


class SftpCli(object):

    def __init__(self):
        self.environment = EnvironmentModel()

    def command_help(self, *args, **kwargs):
        """Displays this help menu."""
        print("Commands available:\n")
        for name in dir(self):
            if not name.startswith("command_"):
                continue
            name_clean = name[len("command_"):]
            print("%s:\n - %s\n" % (name_clean, getattr(self, name).__doc__.strip()))

    def command_setup(self, *args):
        """Defines a new sftp file system configuration or edits an old one with the same id.
        Usage: sftpman setup {options}
        Available {options}:
            --id={unique system identifier}
                You use this to recognize and manage this sftp system.
                It determines what the local mount point is.
                If `--id=example`, the filesystem will be mounted to: `/mnt/sshfs/example`
            --host={host to connect to}
            --port={port to connect to} [default: 22]
            --user={username to authenticate with} [default: current user]
            --mount_opt={option to pass to sshfs} [optional] [can be passed more than once]
                Example: --mount_opt="follow_symlinks" --mount_opt="workaround=rename"
                `sshfs --help` tells you what sshfs options are available
            --mount_point={remote path to mount}
            --auth_method={method}
                Specifies the authentication method.
                Can be `password` or `publickey`. [default: publickey]
            --ssh_key={path to the ssh key to use for authentication}
                Only applies if auth_method is `publickey`.
            --cmd_before_mount={command to run before mounting} [default: /bin/true]
                Allows you to run a custom command every time this system is mounted.
        """
        def usage():
            print(self.command_setup.__doc__)
            sys.exit(1)

        if len(args) == 0:
            usage()

        try:
            # All of these (except mount_opt) map directly to the model properties
            # We allow several `mount_opt` flags and merge their values, before
            # assigning to the `mount_opts` property (which expects a list).
            fields = [
                "id", "host", "port", "user",
                "mount_opt", "mount_point",
                "ssh_key", "cmd_before_mount",
                "auth_method",
            ]
            opts, _ = getopt.getopt(args, "", ["%s=" % s for s in fields])
        except getopt.GetoptError as e:
            sys.stderr.write('Error: %s\n\n' % e)
            usage()

        system = SystemModel()
        mount_opts = []
        for name, value in opts:
            name = name.lstrip('-')
            if not hasattr(system, name):
                continue
            if name == 'mount_opt':
                mount_opts.append(value)
                continue
            setattr(system, name, value)
        system.mount_opts = mount_opts

        is_valid, errors = system.validate()
        if not is_valid:
            sys.stderr.write('Invalid data found:\n')
            for field_name, msg in errors:
                sys.stderr.write(' - %s / %s\n' % (field_name, msg))
            sys.stderr.write('\n')
            usage()
            sys.exit(1)

        system.save(self.environment)
        print('Configuration created.')
        print('You can try mounting now: `sftpman mount %s`' % system.id)

    def command_rm(self, system_id, *system_ids):
        """Removes a system by id.
        Usage: sftpman rm {system_id}..
        For a list of system ids, see `sftpman ls available`.
        """
        # Intentionally reading the first system_id separately,
        # because it's required. The others are optional.
        # This ensures that we'll generate an error if someone tries to call
        # this without the required argument.
        system_ids = (system_id,) + system_ids
        has_failed = False
        for system_id in system_ids:
            try:
                system = SystemModel.create_by_id(system_id, self.environment)
                controller = SystemControllerModel(system, self.environment)
                controller.unmount()
                system.delete(self.environment)
            except SftpException as e:
                sys.stderr.write('Cannot remove %s: %s\n' % (system_id, str(e)))
                has_failed = True
        if has_failed:
            sys.exit(1)

    def command_preflight_check(self):
        """Detects whether we have everything needed to mount sshfs filesystems.
        """
        checks_pass, failures = self.environment.perform_preflight_check()
        if checks_pass:
            print('All checks pass.')
        else:
            sys.stderr.write('Problems encountered:\n')
            for msg in failures:
                sys.stderr.write(' - %s\n' % msg)
            sys.exit(1)

    def command_ls(self, list_what):
        """Lists the available/mounted/unmounted sftp systems.
        Usage: sftpman ls {what}
        Where {what} is one of: available, mounted, unmounted
        """
        if list_what in ('available', 'mounted', 'unmounted'):
            callback = getattr(self.environment, 'get_%s_ids' % list_what)
            lst = callback()
        else:
            lst = []
        if len(lst) != 0:
            print(("\n".join(lst)))

    def command_mount(self, system_id, *system_ids):
        """Mounts the specified sftp system, unless it's already mounted.
        Usage: sftpman mount {id}..
        """
        system_ids = (system_id,) + system_ids
        has_failed = False
        for system_id in system_ids:
            try:
                system = SystemModel.create_by_id(system_id, self.environment)
                controller = SystemControllerModel(system, self.environment)
                controller.mount()
            except SftpConfigException as e:
                sys.stderr.write('Cannot mount %s: %s\n\n' % (system_id, str(e)))
                has_failed = True
            except SftpMountException as e:
                sys.stderr.write('Cannot mount %s!\n\n' % system_id)
                sys.stderr.write('Mount command: \n%s\n\n' % e.mount_cmd)
                sys.stderr.write('Command output: \n%s\n\n' % e.mount_cmd_output)
                has_failed = True
        if has_failed:
            sys.exit(1)

    def command_umount(self, system_id, *system_ids):
        """Unmounts the specified sftp system.
        Usage: sftpman umount {id}..
        """
        system_ids = (system_id,) + system_ids
        has_failed = False
        for system_id in system_ids:
            try:
                system = SystemModel.create_by_id(system_id, self.environment)
                controller = SystemControllerModel(system, self.environment)
                controller.unmount()
            except SftpConfigException as e:
                sys.stderr.write('Cannot unmount %s: %s\n\n' % (system_id, str(e)))
                has_failed = True
        if has_failed:
            sys.exit(1)

    def command_mount_all(self):
        """Mounts all sftp file systems known to sftpman.
        Usage: sftpman mount_all
        """
        has_failed = False
        for system_id in self.environment.get_unmounted_ids():
            try:
                system = SystemModel.create_by_id(system_id, self.environment)
                controller = SystemControllerModel(system, self.environment)
                controller.mount()
            except SftpConfigException as e:
                sys.stderr.write('Cannot mount %s: %s\n\n' % (system_id, str(e)))
                has_failed = True
            except SftpMountException as e:
                sys.stderr.write('Cannot mount %s!\n\n' % system_id)
                sys.stderr.write('Mount command: \n%s\n\n' % e.mount_cmd)
                sys.stderr.write('Command output: \n%s\n\n' % e.mount_cmd_output)
        sys.exit(0 if not has_failed else 1)

    def command_umount_all(self):
        """Unmounts all sftp file systems known to sftpman.
        Usage: sftpman umount_all
        """
        has_failed = False
        for system_id in self.environment.get_mounted_ids():
            try:
                system = SystemModel.create_by_id(system_id, self.environment)
                controller = SystemControllerModel(system, self.environment)
                controller.unmount()
            except SftpConfigException as e:
                sys.stderr.write('Cannot unmount %s: %s\n\n' % (system_id, str(e)))
                has_failed = True
        sys.exit(0 if not has_failed else 1)


def start():
    try:
        command = sys.argv[1]
    except IndexError:
        command = 'help'
    if '--help' in sys.argv:
        command = 'help'
    args = sys.argv[2:]

    instance = SftpCli()
    callback = getattr(instance, "command_%s" % command, None)
    if isinstance(callback, collections.Callable):
        try:
            callback(*args)
        except TypeError as e:
            sys.stderr.write('Bad call for %s: %s' % (command, str(e)))
    else:
        instance.command_help()



# -*- coding: utf-8 -*-

import os
import re
from .helper import json, shell_exec, mkdir_p, rmdir, kill_pid, which
from .exception import SftpConfigException, SftpMountException


class EnvironmentModel(object):
    """A configuration object that represents the environment
    with which we're now working.
    """

    def __init__(self):
        self.mount_path_base = '/mnt/sshfs/'
        cfg_home = os.getenv('XDG_CONFIG_HOME', os.path.expanduser('~/.config'))
        self.config_path_base = "%s/" % os.path.join(cfg_home, 'sftpman')
        self.config_path_mounts = '%smounts/' % self.config_path_base

    def get_system_config_path(self, system_id):
        return '%s%s.json' % (self.config_path_mounts, system_id)

    def get_system_mount_dest(self, system_id):
        """The local path where the system will be mounted."""
        return '%s%s' % (self.mount_path_base, system_id)

    def is_mounted(self, system_id):
        return system_id in self.get_mounted_ids()

    def get_pid_by_system_id(self, system_id):
        # Matching in `{PID} blah blah {mount_dest}`
        mount_dest = self.get_system_mount_dest(system_id)
        regex = re.compile("^(\d+)\s+(?:.+?)\s%s$" % re.escape(mount_dest))

        processes = shell_exec("pgrep --list-full sshfs")
        for line in processes.split("\n"):
            match_object = regex.match(line)
            if match_object is not None:
                return int(match_object.group(1))
        return None

    def get_available_ids(self):
        if not os.path.exists(self.config_path_mounts):
            return []
        cfg_files = shell_exec('ls %s' % self.config_path_mounts).strip().split('\n')
        return [file_name[0:-5] for file_name in cfg_files if file_name.endswith('.json')]

    def get_mounted_ids(self):
        # Looking for /mnt/sshfs/{id} in output that looks like this:
        # user@host:/remote/path on /mnt/sshfs/id type fuse.sshfs ...
        # "mount -l -t fuse.sshfs" cannot be used, as it requires root privileges
        mounted = shell_exec('mount -l')
        regex = re.compile(' %s(.+?) type fuse\.sshfs' % self.mount_path_base)
        return regex.findall(mounted)

    def get_unmounted_ids(self):
        ids_mounted = self.get_mounted_ids()
        return [id for id in self.get_available_ids() if id not in ids_mounted]

    def perform_preflight_check(self):
        """Performs checks to see if we have everything needed to mount
        sshfs filesystems.
        :return: two-tuple (boolean checks_pass, list failure messages)
        """
        failures = []
        if not os.access(self.mount_path_base, os.W_OK):
            msg = ("Mount path `{path}` doesn't exist or is not writable"
                   " by the current user.")
            failures.append(msg.format(
                path = self.mount_path_base
            ))
        if which('sshfs') is None:
            msg = ("SSHFS (http://fuse.sourceforge.net/sshfs.html)"
                   " does not seem to be installed.")
            failures.append(msg)

        return len(failures) == 0, failures


class SystemModel(object):
    """Represents a system (mount point) that sftpman manages."""

    PORT_RANGE_MIN = 0
    PORT_RANGE_MAX = 65535

    SSH_PORT_DEFAULT = 22

    AUTH_METHOD_PUBLIC_KEY = 'publickey'
    AUTH_METHOD_PASSWORD = 'password'

    # libfuse (>=3.0.0) dropped support for big_writes
    UNSUPPORTED_MOUNT_OPTS = ['big_writes']

    def __init__(self, **kwargs):
        self.id = kwargs.get('id', None)
        self.host = kwargs.get('host', None)
        self.port = kwargs.get('port', SystemModel.SSH_PORT_DEFAULT)
        self.user = kwargs.get('user', None)
        self.mount_opts = list(kwargs.get('mountOptions', []))
        self.mount_opts = [v for v in self.mount_opts if v not in SystemModel.UNSUPPORTED_MOUNT_OPTS]
        self.mount_point = kwargs.get('mountPoint', None)
        self.auth_method = kwargs.get('authType', self.AUTH_METHOD_PUBLIC_KEY)
        self.ssh_key = kwargs.get('sshKey', None)
        self.cmd_before_mount = kwargs.get('beforeMount', 'true')

    def _set_port(self, value):
        self._port = int(value)

    port = property(lambda self: self._port, _set_port)

    def _set_ssh_key(self, value):
        if value is not None:
            value = os.path.expanduser(value)
        self._ssh_key = value

    ssh_key = property(lambda self: self._ssh_key, _set_ssh_key)

    def validate(self):
        def is_valid_id(value):
            if value is None:
                return False
            if value in ('.', '..'):
                return False
            return re.compile('^[a-zA-Z0-9\.\-_@]+$').match(value) is not None

        def is_alphanumeric(value):
            if value is None:
                return False
            # Well, not really alphanumeric, but close enough to call it that
            return re.compile('^[a-zA-Z0-9\.\-]+$').match(value) is not None

        def is_valid_username(value):
            if value is None:
                return False
            return re.compile('^[a-zA-Z0-9\.\-\@]+$').match(value) is not None

        def is_valid_path(value):
            if value is None:
                return False
            return re.compile('^/(([a-zA-Z0-9\.\-_]+)/?)*?$').match(value) is not None

        errors = []
        if not is_valid_id(self.id):
            errors.append(('id', 'IDs can only contain letters, digits, dot, @, dash and underscore.'))
        if not is_alphanumeric(self.host):
            errors.append(('host', 'Hosts can only contain letters, digits, dot and dash.'))
        if not is_valid_path(self.mount_point):
            errors.append(('mount_point', 'Invalid remote mount point.'))
        if self.auth_method not in (self.AUTH_METHOD_PUBLIC_KEY, self.AUTH_METHOD_PASSWORD):
            errors.append(('auth_method', 'Unknown auth type.'))
        else:
            if self.auth_method == self.AUTH_METHOD_PUBLIC_KEY:
                if not os.path.exists(self.ssh_key):
                    errors.append(('ssh_key', 'Invalid ssh key path.'))
        if not is_valid_username(self.user):
            errors.append(('user', 'Usernames can only contain letters, at signs and digits.'))
        if not(self.PORT_RANGE_MIN < self.port <= self.PORT_RANGE_MAX):
            msg = 'Ports need to be numbers between %d and %d.' % (
                self.PORT_RANGE_MIN,
                self.PORT_RANGE_MAX,
            )
            errors.append(('port', msg))
        if not isinstance(self.mount_opts, list):
            errors.append(('mount_opts', 'Bad options received.'))
        if not isinstance(self.cmd_before_mount, str) or \
           self.cmd_before_mount == '':
            errors.append(('cmd_before_mount', 'Invalid before mount command.'))

        return (len(errors) == 0, errors)

    def export(self):
        """Serializes to JSON."""
        fields = ['id', 'host', 'port', 'user']
        out = {}
        for field in fields:
            out[field] = getattr(self, field, None)
        out['mountOptions'] = self.mount_opts
        out['mountPoint'] = self.mount_point
        out['beforeMount'] = self.cmd_before_mount
        out['authType'] = self.auth_method
        out['sshKey'] = self.ssh_key
        return json.dumps(out, indent=4)

    def save(self, environment):
        path = environment.get_system_config_path(self.id)
        mkdir_p(os.path.dirname(path))
        with open(path, 'w') as f:
            f.write(self.export())

    def delete(self, environment):
        path = environment.get_system_config_path(self.id)
        os.unlink(path)

    @staticmethod
    def create_by_id(id, environment):
        path = environment.get_system_config_path(id)
        return SystemModel.create_from_file(path)

    @staticmethod
    def create_from_file(path):
        try:
            with open(path) as f:
                config = json.loads(f.read())
                return SystemModel(**config)
        except (ValueError, IOError) as e:
            msg = 'Failed finding or parsing config at %s.'
            raise SftpConfigException(msg % path, e)


class SystemControllerModel(object):
    """Controls a given system within the environment.
    The controller manages mounting, unmounting, cleaning up, etc.
    """

    SIGNAL_SIGTERM = 15
    SIGNAL_SIGKILL = 9

    #: Time to wait when unmounting before forcefully killing the mount process
    KILL_WAIT_TIME_SECONDS = 2

    #: Time to wait for sshfs (ssh) to establish a connection
    SSH_CONNECT_TIMEOUT = 8

    def __init__(self, system, environment):
        self.system = system
        self.environment = environment

    @property
    def mounted(self):
        return self.environment.is_mounted(self.system.id)

    @property
    def mount_point_local(self):
        return self.environment.get_system_mount_dest(self.system.id)

    @property
    def mount_point_remote(self):
        return self.system.mount_point

    def _mount_point_local_create(self):
        """Ensures the mount location exists, so we can start using it."""

        # Ensure nothing's mounted there right now..
        shell_exec('fusermount -u %s' % self.mount_point_local)

        # Ensure the directory path exists
        mkdir_p(self.mount_point_local)

    def _mount_point_local_delete(self):
        rmdir(self.mount_point_local)

    def mount(self):
        """Mounts the sftp system if it's not already mounted."""
        if self.mounted:
            return

        self._mount_point_local_create()

        if len(self.system.mount_opts) == 0:
            sshfs_opts = ""
        else:
            sshfs_opts = " -o %s" % " -o ".join(self.system.mount_opts)

        if self.system.auth_method == self.system.AUTH_METHOD_PUBLIC_KEY:
            ssh_opts = '-o PreferredAuthentications=publickey -i %s' % self.system.ssh_key
        else:
            ssh_opts = '-o PreferredAuthentications=password'

        cmd = ("{cmd_before_mount} &&"
               " sshfs -o ssh_command="
               "'ssh -o ConnectTimeout={timeout} -p {port} {ssh_opts}'"
               " {sshfs_opts} {user}@{host}:{remote_path} {local_path}")
        cmd = cmd.format(
            cmd_before_mount = self.system.cmd_before_mount,
            timeout = self.SSH_CONNECT_TIMEOUT,
            port = self.system.port,
            ssh_opts = ssh_opts,
            sshfs_opts = sshfs_opts,
            host = self.system.host,
            user = self.system.user,
            remote_path = self.mount_point_remote,
            local_path = self.mount_point_local,
        )

        output = shell_exec(cmd).strip()

        if not self.mounted:
            # Clean up the directory tree
            self._mount_point_local_delete()
            if output == '':
                output = 'Mounting failed for a reason unknown to sftpman.'
            raise SftpMountException(cmd, output)

    def unmount(self):
        """Unmounts the sftp system if it's currently mounted."""
        if not self.mounted:
            return

        # Try to unmount properly.
        cmd = 'fusermount -u %s' % self.mount_point_local
        shell_exec(cmd)

        # The filesystem is probably still in use.
        # kill sshfs and re-run this same command (which will work then).
        if self.mounted:
            self._kill()
            shell_exec(cmd)

        self._mount_point_local_delete()

    def _kill(self):
        pid = self.environment.get_pid_by_system_id(self.system.id)
        if pid is None:
            return
        kill_pid(pid, SystemControllerModel.SIGNAL_SIGTERM)

        from time import sleep
        sleep(SystemControllerModel.KILL_WAIT_TIME_SECONDS)

        pid = self.environment.get_pid_by_system_id(self.system.id)
        if pid is not None:
            kill_pid(pid, SystemControllerModel.SIGNAL_SIGKILL)



"""
SftpMan
=======

SftpMan is a command-line application (with a GUI frontend packaged separately) that makes it easy to setup and mount SSHFS/SFTP file systems.

It allows you to define all your SFTP systems and easily mount/unmount them.

A GTK frontend is available, named sftpman-gtk.
"""

from setuptools import setup

setup(
    name = "sftpman",
    version = '1.0.2',
    description = "A command-line application that helps you mount SFTP file systems.",
    long_description = __doc__,
    author = "Slavi Pantaleev",
    author_email = "slavi@devture.com",
    url = "https://github.com/spantaleev/sftpman",
    keywords = ["sftp", "ssh", "sshfs"],
    license = "GPL v3",
    packages = ['sftpman'],
    entry_points="""
    [console_scripts]
    sftpman = sftpman.launcher:main
    """,
    zip_safe = False,
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Unix",
        "Topic :: Communications :: File Sharing",
        "Topic :: Internet",
        "Topic :: Internet :: File Transfer Protocol (FTP)",
        "Topic :: System :: Networking",
        "Topic :: Utilities"
    ]
)
